(*==========================================================================
*
*  Copyright (C) 1995-1999 Microsoft Corporation. All Rights Reserved.
*  Ported by Vadim Bodrov, TMT Development Corp.
*
*  File:       donuts.c
*  Content:    Shoot-em-up game
*
*
***************************************************************************)

{$define USE_DSOUND}
{$i donuts.inc}
{$r donuts.res}

var
    lpDD:            PPDirectDraw7;
    lpFrontBuffer:   PPDirectDrawSurface7;
    lpBackBuffer:    PPDirectDrawSurface7;
    lpDonut:         PPDirectDrawSurface7;
    lpPyramid:       PPDirectDrawSurface7;
    lpCube:          PPDirectDrawSurface7;
    lpSphere:        PPDirectDrawSurface7;
    lpShip:          PPDirectDrawSurface7;
    lpNum:           PPDirectDrawSurface7;
    lpArtPalette:    PPDirectDrawPalette;
    lpSplashPalette: PPDirectDrawPalette;
    bSoundEnabled:   Boolean;
    bPlayIdle:       Boolean := FALSE;
    bPlayBuzz:       Boolean := FALSE;
    bPlayRev:        Boolean := FALSE;
    lastThrust:      Boolean := FALSE;
    lastShield:      Boolean := FALSE;
    showDelay:       Longint := 0;
    hAccel:          HACCEL;
    bShowFrameCount: Boolean := TRUE;
    bIsActive:       Boolean;
    bMouseVisible:   Boolean;
    dwFrameCount:    DWORD;
    dwFrameTime:     DWORD;
    dwFrames:        DWORD;
    dwFramesLast:    DWORD;
    bUseEmulation:   Boolean;
    bTest:           Boolean := FALSE;
    bStress:         Boolean := FALSE;
    dwTransType:     DWORD;
    SPalette:        array[0..255] of TRGBQuad;
    lastTickCount:   DWORD;
    score:           Longint;
    ProgramState:    Longint;
    level:           Longint;
    restCount:       Longint;
    dwFillColor:     DWORD;
    bSpecialEffects: Boolean := FALSE;
    iForceErase:     Longint := 0;
    ShowLevelCount:  DWORD   := 3000;
    bWantSound:      Boolean := TRUE; //global hack to turn off sound


function getint(var p: PChar; def: Longint): Longint;
var
   i: Longint := 0;
begin
    while (IS_SPACE(p^)) do
        inc(p);

    if (not IS_NUM (p^)) then
    begin
        Result := def;
        exit;
    end;

    while (IS_NUM(p^)) do
    begin
        i := i * 10 + Ord(p^) - Ord('0');
        inc(p);
    end;

    while (IS_SPACE(p^)) do
        inc(p);

    Result := i;
end;


var
    DL: TDblNode;             // Display List

{$ifdef USE_DSOUND}
    lpDS:              PPDirectSound;
    hsoBeginLevel:     PSndObj;
    hsoEngineIdle:     PSndObj;
    hsoEngineRev:      PSndObj;
    hsoSkidToStop:     PSndObj;
    hsoShieldBuzz:     PSndObj;
    hsoShipExplode:    PSndObj;
    hsoFireBullet:     PSndObj;
    hsoShipBounce:     PSndObj;
    hsoDonutExplode:   PSndObj;
    hsoPyramidExplode: PSndObj;
    hsoCubeExplode:    PSndObj;
    hsoSphereExplode:  PSndObj;
{$endif}

{$ifdef USE_DSOUND}
procedure InitializeSound;
begin
    if (not bWantSound) then
        exit; // out of here
    bSoundEnabled := FALSE;
    if (SUCCEEDED(DirectSoundCreate(nil, lpDS, nil))) then
    begin
        if (SUCCEEDED(lpDS^^.SetCooperativeLevel^(lpDS, hWndMain, DSSCL_NORMAL))) then
        begin
            hsoBeginLevel     := SndObjCreate(lpDS, 'BeginLevel',      1);
            hsoEngineIdle     := SndObjCreate(lpDS, 'EngineIdle',      1);
            hsoEngineRev      := SndObjCreate(lpDS, 'EngineRev',       1);
            hsoSkidToStop     := SndObjCreate(lpDS, 'SkidToStop',      1);
            hsoShieldBuzz     := SndObjCreate(lpDS, 'ShieldBuzz',      1);
            hsoShipExplode    := SndObjCreate(lpDS, 'ShipExplode',     1);
            hsoFireBullet     := SndObjCreate(lpDS, 'Gunfire',        25);
            hsoShipBounce     := SndObjCreate(lpDS, 'ShipBounce',      4);
            hsoDonutExplode   := SndObjCreate(lpDS, 'DonutExplode',   10);
            hsoPyramidExplode := SndObjCreate(lpDS, 'PyramidExplode', 12);
            hsoCubeExplode    := SndObjCreate(lpDS, 'CubeExplode',    15);
            hsoSphereExplode  := SndObjCreate(lpDS, 'SphereExplode',  10);
            bSoundEnabled     := TRUE;

           //{$ifdef USE_DSOUND   this should be dead code Josephc}
           if ( bPlayIdle ) then
                SndObjPlay(hsoEngineIdle, DSBPLAY_LOOPING);

           if ( bPlayBuzz ) then
                SndObjPlay(hsoShieldBuzz, DSBPLAY_LOOPING);

           if ( bPlayRev ) then
                SndObjPlay(hsoEngineRev, DSBPLAY_LOOPING);
           //{$endif}
        end else
        begin
            lpDS^^.Release^(lpDS);
            lpDS := nil;
        end;
    end;
end;

procedure DestroySound;
begin
    if (not bWantSound) then
        exit; //No work to be done
    bSoundEnabled := FALSE;
    if (lpDS <> nil) then
    begin
        SndObjDestroy(hsoBeginLevel);
        hsoBeginLevel := nil;
        SndObjDestroy(hsoEngineIdle);
        hsoEngineIdle := nil;
        SndObjDestroy(hsoEngineRev);
        hsoEngineRev := nil;
        SndObjDestroy(hsoSkidToStop);
        hsoSkidToStop := nil;
        SndObjDestroy(hsoShieldBuzz);
        hsoShieldBuzz := nil;
        SndObjDestroy(hsoShipExplode);
        hsoShipExplode := nil;
        SndObjDestroy(hsoFireBullet);
        hsoFireBullet := nil;
        SndObjDestroy(hsoShipBounce);
        hsoShipBounce := nil;
        SndObjDestroy(hsoDonutExplode);
        hsoDonutExplode := nil;
        SndObjDestroy(hsoPyramidExplode);
        hsoPyramidExplode := nil;
        SndObjDestroy(hsoCubeExplode);
        hsoCubeExplode := nil;
        SndObjDestroy(hsoSphereExplode);
        hsoSphereExplode := nil;
        lpDS^^.Release^(lpDS);
        lpDS := nil;
    end;
end;
{$endif}


procedure DeleteFromList( this: PDblNode );
begin
    this^.next^.prev := this^.prev;
    this^.prev^.next := this^.next;
    LocalFree( HLOCAL(this) );
end;


function randDouble( low, high: Double ): Double;
var
    num, range: Double;

begin
    range := high - low;
    num := range * (Random(32768) / 32768);
    Result := num + low;
end;

procedure linkObject( new: PDBlNode );
begin
    new^.next := DL.next;
    new^.prev := @DL;
    DL.next^.prev := new;
    DL.next := new;
end;


procedure linkLastObject( new: PDBlNode );
begin
    new^.prev := DL.prev;
    new^.next := @DL;
    DL.prev^.next := new;
    DL.prev := new;
end;


procedure addObject( objtype: SHORT; x, y, vx, vy: Double );
var
    new: PDblNode;
begin
    new := PDblNode(LocalAlloc( LPTR, sizeof( TDblNode ) ));
    if new = nil then
        exit;

    new^.objtype := objtype;
    case objtype of
        OBJ_DONUT:
        begin
            if ( x < 0.0) then // no position specified?
            begin
                new^.posx := randDouble( 0.0, MAX_DONUT_X );
                new^.posy := randDouble( 0.0, MAX_DONUT_Y );
            end else
            begin
                new^.posx := x;
                new^.posy := y;
            end;
            new^.velx := randDouble( -50.0/1000.0, 50.0/1000.0 );
            new^.vely := randDouble( -50.0/1000.0, 50.0/1000.0 );
            new^.frame := randDouble( 0, 30 );
            new^.delay := 30.0*randDouble( 0.1, 0.4 )/1000.0;
            new^.surf := lpDonut;
            linkObject( new );
        end;
        OBJ_PYRAMID:
        begin
            if ( x < 0) then // no position specified?
            begin
                new^.posx := randDouble( 0.0, MAX_PYRAMID_X );
                new^.posy := randDouble( 0.0, MAX_PYRAMID_Y );
            end else
            begin
                new^.posx := x;
                new^.posy := y;
            end;
            new^.velx := 1.5*randDouble( -50.0/1000.0, 50.0/1000.0 );
            new^.vely := 1.5*randDouble( -50.0/1000.0, 50.0/1000.0 );
            new^.frame := randDouble( 0, 30 );
            new^.delay := 40.0*randDouble( 0.3, 1.0 )/1000.0;
            new^.surf := lpPyramid;
            linkObject( new );
        end;
        OBJ_SPHERE:
        begin
            if ( x < 0) then // no position specified?
            begin
                new^.posx := randDouble( 0.0, MAX_SPHERE_X );
                new^.posy := randDouble( 0.0, MAX_SPHERE_Y );
            end else
            begin
                new^.posx := x;
                new^.posy := y;
            end;
            new^.velx := 3.0*randDouble( -50.0/1000.0, 50.0/1000.0 );
            new^.vely := 3.0*randDouble( -50.0/1000.0, 50.0/1000.0 );
            new^.frame := randDouble( 0, 30 );
            new^.delay := 40.0*randDouble( 1.5, 2.0 )/1000.0;
            new^.surf := lpSphere;
            linkObject( new );
        end;
        OBJ_CUBE:
        begin
            if ( x < 0) then // no position specified?
            begin
                new^.posx := randDouble( 0.0, MAX_CUBE_X );
                new^.posy := randDouble( 0.0, MAX_CUBE_Y );
            end else
            begin
                new^.posx := x;
                new^.posy := y;
            end;
            new^.velx := 4.0*randDouble( -50.0/1000.0, 50.0/1000.0 );
            new^.vely := 4.0*randDouble( -50.0/1000.0, 50.0/1000.0 );
            new^.frame := randDouble( 0, 30 );
            new^.delay := 40.0*randDouble( 0.8, 2.0 )/1000.0;
            new^.surf := lpCube;
            linkObject( new );
        end;
        OBJ_BULLET:
        begin
            new^.posx := x;
            new^.posy := y;
            new^.velx := vx;
            new^.vely := vy;
            new^.frame := 0.0;
            new^.delay := 1.0;
            new^.surf := lpNum;
            linkObject( new );
        end;
    end;
end;

procedure initShip( delay: Boolean );
begin
    DL.posx := (ScreenX div 2-16);       // center the ship
    DL.posy := (ScreenY div 2-16);
    DL.frame := 0.0;
    if bTest then
    begin
        DL.velx := 0.25;
        DL.vely := 0.5;
    end else
    begin
        DL.velx := 0.0;                  // not moving
        DL.vely := 0.0;
    end;
    if ( not bTest) and ( delay ) then
        showDelay := DEF_SHOW_DELAY;
end;

procedure initLevel( level: Longint );
var
    i: Longint;
begin

    // clear any stray bullets out of the display list
    while( DL.next <> @DL ) do
    begin
        DeleteFromList( DL.next );
    end;
    for i := 0 to (2*level-2) do
    begin
        addObject( OBJ_DONUT, -1.0, -1.0, -1.0, -1.0 );
    end;
    initShip(TRUE);
end;


procedure setup_game;
begin
    restCount := GetTickCount;
    inc(level);
    initLevel(level);
    // set the palette
    lpFrontBuffer^^.SetPalette^( lpFrontBuffer, lpArtPalette );
end;

(*
 *
 *  AppPause
 *
 *)
procedure AppPause;
begin
    lpDD^^.FlipToGDISurface^(lpDD);
    DrawMenuBar(hWndMain);
    RedrawWindow(hWndMain, nil, 0, RDW_FRAME);
end;

(*
*
*  AppUnpause
*
*  Reset the various time counters so the donuts don't suddenly
*  jump halfways across the screen and so the frame rate remains accurate.
*
*)
procedure AppUnpause;
begin
    iForceErase := 2;
    dwFrameTime := timeGetTime;
    lastTickCount := dwFrameTime;
end;

(*
*  CheckOneMenuItem
*
*  Helper function that checks a single item in a menu.
*)
procedure CheckOneMenuItem(hmenu: HMENU; idc: UINT; fCheck: Boolean);
var
    uCheck: UINT;
begin
    if fCheck then
        uCheck := MF_BYCOMMAND or MF_CHECKED
    else
        uCheck := MF_BYCOMMAND or MF_UNCHECKED;
    CheckMenuItem(hmenu, idc, uCheck);
end;

(*
*  CheckMenuItems
*
*  Sync the menu checkmarks with our internal variables
*)
procedure CheckMenuItems( wnd: HWND );
var
    menu: HMENU;
begin
    menu := GetMenu(wnd);
    CheckOneMenuItem(menu, IDC_TRAILS, bSpecialEffects);
{$ifdef USE_DSOUND}
    CheckOneMenuItem(menu, IDC_AUDIO, bWantSound and bSoundEnabled);
{$endif}
    CheckOneMenuItem(menu, IDC_FRAMERATE, bShowFrameCount);
end;

procedure DestroyGame;
begin
end;

(*
* MainWndproc
*
* Callback for all Windows messages
*)
function MainWndproc conv arg_stdcall ( Wnd: HWND; message: UINT; wParam: WPARAM; lParam: LPARAM ): Longint;
var
    ps:  TPaintStruct;
    dc:  HDC;
    cmd: UINT;
begin
    case message  of
        WM_ACTIVATEAPP:
        begin
            bIsActive := Boolean(wParam);
            if bIsActive then
            begin
                bMouseVisible := FALSE;
                lastTickCount := GetTickCount;
                bSpecialEffects := FALSE;
                // we are active, need to reacquire the keyboard
                ReacquireInput;
            end else
            begin
                bMouseVisible := TRUE;
                // DirectInput automatically unacquires for us in FOREGROUND mode
            end;
        end;

        WM_CREATE:
            exit;

        WM_SETCURSOR:
        begin
            if ( not bMouseVisible ) then
                SetCursor(0)
            else
                SetCursor(LoadCursor( 0, IDC_ARROW ));
            Result := LONGTRUE;
            exit;
        end;

        WM_COMMAND:
        begin
            cmd := GET_WM_COMMAND_ID(wParam, lParam);

            if (cmd >= IDC_DEVICES) and (cmd < (IDC_DEVICES + 100)) then
                PickInputDevice(cmd - IDC_DEVICES)
            else
            begin
                case cmd of
                    IDC_FRAMERATE:
                    begin
                        bShowFrameCount := not bShowFrameCount;
                        if bShowFrameCount then
                        begin
                            dwFrameCount := 0;
                            dwFrameTime := timeGetTime;
                        end;
                    end;

                    IDC_STARTGAME:
                    begin
                        if ( ProgramState = PS_SPLASH ) then
                        begin
                            ProgramState := PS_BEGINREST;
                            setup_game;
                        end;
                    end;

                    IDC_QUIT:
                    begin
                        PostMessage( Wnd, WM_CLOSE, 0, 0 );
                        Result := 0;
                        exit;
                    end;

                    IDC_AUDIO:
                    begin
                    {$ifdef USE_DSOUND}
                        if (bWantSound) then
                        begin
                            if ( bSoundEnabled ) then
                                DestroySound
                            else
                                InitializeSound;
                        end;
                    {$endif}
                    end;

                    IDC_TRAILS:
                    begin
                        bSpecialEffects := not bSpecialEffects;
                    end;
                end; { case }
            end; { if }
        end;

        WM_INITMENU:
        begin
            CheckMenuItems(hWndMain);
        end;

        WM_ERASEBKGND:
        begin
            Result := 1;
            exit;
        end;

        WM_PAINT:
        begin
            dc := BeginPaint( Wnd, ps );
            EndPaint( Wnd, ps );
            Result := 1;
            exit;
        end;

        WM_DESTROY:
        begin
            CleanupInput;
            DestroyGame;
            PostQuitMessage( 0 );
        end;

        WM_ENTERMENULOOP:
        begin
            AppPause;
        end;

        WM_EXITMENULOOP:
        begin
            AppUnpause;
        end;

    end;
    Result := DefWindowProc(Wnd, message, wParam, lParam);

end; (* MainWndproc *)

  (*
  * initApplication
  *
  * Do that Windows initialization stuff...
  *)
function initApplication( nCmdShow: Longint ): Boolean;
var
    wc: TWndClass;
    rc: LONGBOOL;
begin
    wc.style := CS_DBLCLKS;
    wc.lpfnWndProc := @MainWndproc;
    wc.cbClsExtra := 0;
    wc.cbWndExtra := 0;
    wc.hInstance := hInstance;
    wc.hIcon := LoadIcon( hInstance, MAKEINTRESOURCE(DONUTS_ICON));
    wc.hCursor := LoadCursor( 0, IDC_ARROW );
    wc.hbrBackground := GetStockObject( BLACK_BRUSH );
    wc.lpszMenuName :=  MAKEINTRESOURCE(DONUTS_MENU);
    wc.lpszClassName := 'DonutsClass';
    rc := RegisterClass( wc );

    if ( rc = LONGFALSE ) then
    begin
        Result := FALSE;
        exit;
    end;

    hAccel := LoadAccelerators(hInstance, MAKEINTRESOURCE(DONUTS_ACCEL));
    if ( hAccel = 0 ) then
    begin
        Result := FALSE;
        exit;
    end;

    hWndMain := CreateWindowEx(0,  // WS_EX_TOPMOST,
        'DonutsClass',
        'Donuts',
        WS_VISIBLE or // so we don't have to call ShowWindow
        WS_POPUP or   // non-app window
        WS_CAPTION or // so our menu doesn't look ultra-goofy
        WS_SYSMENU,   // so we get an icon in the tray
        0,
        0,
        GetSystemMetrics(SM_CXSCREEN),
        GetSystemMetrics(SM_CYSCREEN),
        0,
        0,
        hInstance,
        nil );

    if ( hWndMain = 0 ) then
    begin
        Result := FALSE;
        exit;
    end;

    UpdateWindow( hWndMain );

{$ifdef USE_DSOUND}
      (*
      *  If sound is globally disabled, then disable the sound menu.
      *)
      if (not bWantSound)  then
          EnableMenuItem(GetMenu(hWndMain), IDC_AUDIO, MF_BYCOMMAND or MF_DISABLED or MF_GRAYED);
{$endif}

    Result := TRUE;

end; (* initApplication *)

function CleanupAndExit( err: PChar ): Boolean;
begin
    // make the cursor visible
    SetCursor(LoadCursor( 0, IDC_ARROW ));
    bMouseVisible := TRUE;

    if ( lpDonut <> nil ) then
        lpDonut^^.Release^( lpDonut );

    if ( lpPyramid <> nil ) then
        lpPyramid^^.Release^( lpPyramid );

    if ( lpCube <> nil ) then
        lpCube^^.Release^( lpCube );

    if ( lpSphere <> nil ) then
        lpSphere^^.Release^( lpSphere );

    if ( lpShip <> nil ) then
        lpShip^^.Release^( lpShip );

    if ( lpNum <> nil ) then
        lpNum^^.Release^( lpNum );

    if ( lpFrontBuffer <> nil ) then
        lpFrontBuffer^^.Release^( lpFrontBuffer );

    if ( lpArtPalette <> nil ) then
        lpArtPalette^^.Release^( lpArtPalette );

    if ( lpSplashPalette <> nil ) then
        lpSplashPalette^^.Release^( lpSplashPalette );

    if ( lpDD <> nil ) then
        lpDD^^.Release^( lpDD );

    // clean up DirectInput objects
    CleanupInput;

    //
    // warn user if there is one
    //

    if ( not bStress ) then
    begin
        MessageBox( hWndMain, err, 'ERROR', MB_OK );
    end;
    Result := FALSE;
end;

function RestoreSurfaces: Boolean;
var
    ddrval: HResult;
    hbm:    HBITMAP;
begin
    ddrval := lpFrontBuffer^^.Restore^(lpFrontBuffer);
    if ( ddrval <> DD_OK ) then
    begin
        Result := FALSE;
        exit;
    end;
    ddrval := lpDonut^^.Restore^(lpDonut);
    if ( ddrval <> DD_OK ) then
    begin
        Result := FALSE;
        exit;
    end;
    ddrval := lpPyramid^^.Restore^(lpPyramid);
    if ( ddrval <> DD_OK ) then
    begin
        Result := FALSE;
        exit;
    end;
    ddrval := lpCube^^.Restore^(lpCube);
    if ( ddrval <> DD_OK ) then
    begin
        Result := FALSE;
        exit;
    end;
    ddrval := lpSphere^^.Restore^(lpSphere);
    if ( ddrval <> DD_OK ) then
    begin
        Result := FALSE;
        exit;
    end;
    ddrval := lpShip^^.Restore^(lpShip);
    if ( ddrval <> DD_OK ) then
    begin
        Result := FALSE;
        exit;
    end;
    ddrval := lpNum^^.Restore^(lpNum);
    if ( ddrval <> DD_OK ) then
    begin
        Result := FALSE;
        exit;
    end;

    // Create and set the palette for the splash bitmap
    lpSplashPalette := DDLoadPalette( lpDD, 'SPLASH' );
    if ( lpSplashPalette = nil ) then
    begin
        Result := CleanupAndExit('DDLoadPalette SPLASH');
        exit;
    end;

    // Create and set the palette for the art bitmap
    lpArtPalette := DDLoadPalette( lpDD, 'DONUTS8' );
    if ( lpArtPalette = nil ) then
    begin
        Result := CleanupAndExit('DDLoadPalette DONUTS');
        exit;
    end;

    // set the palette before loading the art
    lpFrontBuffer^^.SetPalette^( lpFrontBuffer, lpArtPalette );

    hbm := LoadImage(GetModuleHandle(nil), 'DONUTS8', IMAGE_BITMAP, 0, 0, LR_CREATEDIBSECTION );

    if ( hbm = 0 ) then
    begin
        Result := FALSE;
        exit;
    end;

    ddrval := DDCopyBitmap( lpDonut, hbm, 0, 0, 320, 384 );
    if ( ddrval <> DD_OK ) then
    begin
        DeleteObject( hbm );
        Result := FALSE;
        exit;
    end;

    // NOTE: Why are we calling LoadImage again?  StretchBlt (which is
    // called in DDCopyBitmap) does not work properly when performing
    // an 8-bpp to 24- or 32-bpp blt multiple times from the same
    // bitmap.  The workaround is to call LoadImage before each
    // StretchBlt because the first StretchBlt after a LoadImage will
    // work.
    if (ScreenBpp >= 24) then
    begin
        DeleteObject( hbm );
        hbm := LoadImage(GetModuleHandle(nil), 'DONUTS8', IMAGE_BITMAP, 0, 0, LR_CREATEDIBSECTION );

        if ( hbm = 0 ) then
        begin
            Result := FALSE;
            exit;
        end;
    end;

    ddrval := DDCopyBitmap( lpPyramid, hbm, 0, 384, 320, 128 );
    if ( ddrval <> DD_OK ) then
    begin
        DeleteObject( hbm );
        Result := FALSE;
        exit;
    end;

    if (ScreenBpp >= 24) then
    begin
        DeleteObject( hbm );
        hbm := LoadImage(GetModuleHandle(nil), 'DONUTS8', IMAGE_BITMAP, 0, 0, LR_CREATEDIBSECTION );

        if ( hbm = 0 ) then
        begin
            Result := FALSE;
            exit;
        end;
    end;

    ddrval := DDCopyBitmap( lpSphere, hbm, 0, 512, 320, 32 );
    if ( ddrval <> DD_OK ) then
    begin
        DeleteObject( hbm );
        Result := FALSE;
        exit;
    end;

    if (ScreenBpp >= 24) then
    begin
        DeleteObject( hbm );
        hbm := LoadImage(GetModuleHandle(nil), 'DONUTS8', IMAGE_BITMAP, 0, 0, LR_CREATEDIBSECTION );

        if ( hbm = 0 ) then
        begin
            Result := FALSE;
            exit;
        end;
    end;

    ddrval := DDCopyBitmap( lpCube, hbm, 0, 544, 320, 32 );
    if ( ddrval <> DD_OK ) then
    begin
        DeleteObject( hbm );
        Result := FALSE;
        exit;
    end;

    if (ScreenBpp >= 24) then
    begin
        DeleteObject( hbm );
        hbm := LoadImage(GetModuleHandle(nil), 'DONUTS8', IMAGE_BITMAP, 0, 0, LR_CREATEDIBSECTION );

        if ( hbm = 0 ) then
        begin
            Result := FALSE;
            exit;
        end;
    end;

    ddrval := DDCopyBitmap( lpShip, hbm, 0, 576, 320, 256 );
    if ( ddrval <> DD_OK ) then
    begin
        DeleteObject( hbm );
        Result := FALSE;
        exit;
    end;

    if (ScreenBpp >= 24) then
    begin
        DeleteObject( hbm );
        hbm := LoadImage(GetModuleHandle(nil), 'DONUTS8', IMAGE_BITMAP, 0, 0, LR_CREATEDIBSECTION );

        if ( hbm = 0 ) then
        begin
            Result := FALSE;
            exit;
        end;
    end;

    ddrval := DDCopyBitmap( lpNum, hbm, 0, 832, 320, 16 );
    if ( ddrval <> DD_OK ) then
    begin
        DeleteObject( hbm );
        Result := FALSE;
        exit;
    end;

    DeleteObject( hbm );

    // set colorfill colors and color keys according to bitmap contents
    dwFillColor := DDColorMatch( lpDonut, CLR_INVALID );

    DDSetColorKey( lpDonut, CLR_INVALID );
    DDSetColorKey( lpPyramid, CLR_INVALID );
    DDSetColorKey( lpCube, CLR_INVALID );
    DDSetColorKey( lpSphere, CLR_INVALID );
    DDSetColorKey( lpShip, CLR_INVALID );
    DDSetColorKey( lpNum, CLR_INVALID );

    Result := TRUE;
end;

function InitializeGame: Boolean;
var
    ddcaps:     TDDCaps;
    ddrval:     HResult;
    ddsd:       TDDSurfaceDesc2;
    ddscaps:    TDDSCaps2;
{$ifdef NT_HACK}
    DDSurfDesc: TDDSurfaceDesc2;
{$endif}
begin
    score := 0;
    if ( bTest ) then
        ShowLevelCount := 1000;

    if ( bUseEmulation ) then
        ddrval := DirectDrawCreateEx(Pointer(DDCREATE_EMULATIONONLY), lpDD, IID_IDirectDraw7, nil)
    else
        ddrval := DirectDrawCreateEx( nil, lpDD, IID_IDirectDraw7, nil);

    if ( ddrval <> DD_OK ) then
    begin
        Result := CleanupAndExit('DirectDrawCreate Failed!');
        exit;
    end;

    ddrval := lpDD^^.SetCooperativeLevel^( lpDD, hWndMain,
        DDSCL_EXCLUSIVE or DDSCL_FULLSCREEN );
    if ( ddrval <> DD_OK ) then
    begin
        Result := CleanupAndExit('SetCooperativeLevel Failed');
        exit;
    end;

{$ifdef NT_HACK}
    DDSurfDesc.dwSize := sizeof(DDSurfDesc);
    ddrval = lpDD^^.GetDisplayMode^(lpDD, DDSurfDesc);
    if (ddrval = DD_OK) then
          ScreenBpp := DDSurfDesc.ddpfPixelFormat.dwRGBBitCount;
{$endif}


    // set the mode
    ddrval := lpDD^^.SetDisplayMode^( lpDD, ScreenX, ScreenY, ScreenBpp, 0, 0 );
    if ( ddrval <> DD_OK ) then
    begin
        Result := CleanupAndExit('SetDisplayMode Failed!');
        exit;
    end;

    // check the color key hardware capabilities
    dwTransType := DDBLTFAST_SRCCOLORKEY;
    ddcaps.dwSize := sizeof( ddcaps );

    // Create surfaces
    ZeroMemory( @ddsd, sizeof( ddsd ) );
    ddsd.dwSize := sizeof( ddsd );
    ddsd.dwFlags := DDSD_CAPS or DDSD_BACKBUFFERCOUNT;
    ddsd.ddsCaps.dwCaps := DDSCAPS_PRIMARYSURFACE or
        DDSCAPS_FLIP or DDSCAPS_COMPLEX;
    ddsd.dwBackBufferCount := 1;
    ddrval := lpDD^^.CreateSurface^( lpDD, ddsd, lpFrontBuffer, nil );
    if ( ddrval <> DD_OK ) then
    begin
        Result := CleanupAndExit('CreateSurface FrontBuffer Failed!');
        exit;
    end;

    // get a pointer to the back buffer
    ZeroMemory(@ddscaps, sizeof(ddscaps));
    ddscaps.dwCaps := DDSCAPS_BACKBUFFER;
    ddrval := lpFrontBuffer^^.GetAttachedSurface^(
        lpFrontBuffer,
        ddscaps,
        lpBackBuffer );
    if ( ddrval <> DD_OK ) then
    begin
        Result := CleanupAndExit('GetAttachedDurface Failed!');
        exit;
    end;

    ddsd.dwFlags := DDSD_CAPS or DDSD_WIDTH or DDSD_HEIGHT;
    ddsd.ddsCaps.dwCaps := DDSCAPS_OFFSCREENPLAIN;
    ddsd.dwWidth := 320;
    ddsd.dwHeight := 384;
    ddrval := lpDD^^.CreateSurface^( lpDD, ddsd, lpDonut, nil );
    if ( ddrval <> DD_OK ) then
    begin
        Result := CleanupAndExit('CreateSurface lpDonut Failed!');
        exit;
    end;

    ddsd.dwHeight := 128;
    ddrval := lpDD^^.CreateSurface^( lpDD, ddsd, lpPyramid, nil );
    if ( ddrval <> DD_OK ) then
    begin
        Result := CleanupAndExit('CreateSurface lpPyramid Failed!');
        exit;
    end;

    ddsd.dwHeight := 32;
    ddrval := lpDD^^.CreateSurface^( lpDD, ddsd, lpCube, nil );
    if ( ddrval <> DD_OK ) then
    begin
        Result := CleanupAndExit('CreateSurface lpCube Failed!');
        exit;
    end;

    ddsd.dwHeight := 32;
    ddrval := lpDD^^.CreateSurface^( lpDD, ddsd, lpSphere, nil );
    if ( ddrval <> DD_OK ) then
    begin
        Result := CleanupAndExit('CreateSurface lpSphere Failed!');
        exit;
    end;
    // Set the background color fill color

    ddsd.dwHeight := 256;
    ddrval := lpDD^^.CreateSurface^( lpDD, ddsd, lpShip, nil );
    if ( ddrval <> DD_OK ) then
    begin
        Result := CleanupAndExit('CreateSurface lpShip Failed!');
        exit;
    end;

    ddsd.dwHeight := 16;
    ddrval := lpDD^^.CreateSurface^( lpDD, ddsd, lpNum, nil );
    if ( ddrval <> DD_OK ) then
    begin
        Result := CleanupAndExit('CreateSurface lpNum Failed!');
        exit;
    end;

    if ( not RestoreSurfaces ) then
    begin
        Result := CleanupAndExit('RestoreSurfaces Failed!');
        exit;
    end;

    DL.prev := @DL;            // null display list
    DL.next := @DL;
    DL.objtype := OBJ_SHIP;
    DL.surf := lpShip;
    lastTickCount := GetTickCount;

    {$ifdef USE_DSOUND}
      if (bWantSound) then
          InitializeSound;
    {$endif}

    // initialize DirectInput
    if (not InitInput(hInstance, hWndMain)) then
    begin
        Result := CleanupAndExit('DirectInput initialization Failed!');
        exit;
    end;

    if (bTest) then
    begin
        ProgramState := PS_ACTIVE;
        setup_game;
    end else
    begin
        ProgramState := PS_SPLASH;
    end;
    Result := TRUE;
end;

procedure FlipScreen;
var
    ddrval: HResult;
begin
    while( TRUE ) do
    begin
        ddrval := lpFrontBuffer^^.Flip^( lpFrontBuffer, nil, 0 );
        if ( ddrval = DD_OK ) then
            break;
        if ( ddrval <> DDERR_SURFACELOST ) then
        begin
            if( not RestoreSurfaces ) then
                exit;
        end;
        if ( ddrval <> DDERR_WASSTILLDRAWING ) then
            break
    end;
end;

procedure bltSplash;
var
    ddrval: HResult;
    hbm:    HBitmap;
begin
    // set the palette before loading the splash screen
    lpFrontBuffer^^.SetPalette^( lpFrontBuffer, lpSplashPalette );

    hbm := LoadImage( GetModuleHandle( nil ), 'SPLASH', IMAGE_BITMAP, 0, 0, LR_CREATEDIBSECTION );
    if ( hbm = 0 ) then
        exit;

    // if the surface is lost, DDCopyBitmap will fail and the surface will
    // be restored in FlipScreen.
    ddrval := DDCopyBitmap( lpBackBuffer, hbm, 0, 0, 0, 0 );

    DeleteObject( hbm );

    FlipScreen;
end;

procedure EraseScreen;
var
   ddbltfx: TDDBltFX;
   ddrval:  HResult;
begin
    // iForceErase forces us to erase so we can get of stray GDI pixels
    if (iForceErase <> 0) then
        iForceErase -:= 1
    else
    if( bSpecialEffects ) then  // cool looking screen with no colorfill
        exit;

    // Erase the background
    ZeroMemory(@ddbltfx, sizeof(ddbltfx));
    ddbltfx.dwSize := sizeof( ddbltfx );
    ddbltfx.dwFillColor := dwFillColor;
    while( TRUE ) do
    begin
        ddrval := lpBackBuffer^^.Blt^( lpBackBuffer, nil, nil,
            nil, DDBLT_COLORFILL, @ddbltfx );

        if ( ddrval = DD_OK ) then
             exit;

        if( ddrval = DDERR_SURFACELOST ) then
        begin
            if( not RestoreSurfaces ) then
                exit;
        end;
        if( ddrval <> DDERR_WASSTILLDRAWING ) then
            exit;
    end;
end;

procedure bltScore( num: PChar; x, y: Longint );
var
    c:      PChar;
    src:    TRect;
    i:      Longint;
    ddrval: HResult;
begin
    c := num;
    while (Char(c^) <> #0) do
    begin
        while( TRUE ) do
        begin
            i := Byte(c^) - Ord('0');
            src.left := i*16;
            src.top := 0;
            src.right := src.left + 16;
            src.bottom := src.top + 16;
            ddrval := lpBackBuffer^^.BltFast^( lpBackBuffer, x, y, lpNum, src, dwTransType );
            if ( ddrval = DD_OK ) then
                 break;

            if ( ddrval = DDERR_SURFACELOST ) then
            begin
                if ( not RestoreSurfaces ) then
                    exit;
            end;
            if ( ddrval <> DDERR_WASSTILLDRAWING ) then
                 exit;
        end;
        x +:= 16;
        inc(c);
    end;
end;

procedure DisplayLevel;
var
    buf: array[0..9] of Byte;
begin
    EraseScreen;
    buf[0] := 10 + Ord('0');
    buf[1] := 11 + Ord('0');
    buf[2] := 12 + Ord('0');
    buf[3] := 11 + Ord('0');
    buf[4] := 10 + Ord('0');
    buf[5] := 0;
    bltScore( PChar(@buf[0]), ScreenX div 2-64, ScreenY div 2-8 );
    buf[0] := level div 100 + Ord('0');
    buf[1] := level div 10  + Ord('0');
    buf[2] := level mod 10  + Ord('0');
    buf[3] := 0;
    bltScore( PChar(@buf), ScreenX div 2+22, ScreenY div 2-8 );
    FlipScreen;
end;

{$ifdef USE_DSOUND}
//
// play a sound, but first set the panning according to where the
// object is on the screen.  fake 3D sound.
//
procedure playPanned(hSO: PSNDOBJ; uobject: PDblNode);
var
    pDSB: PPDirectSoundBuffer;
begin
    pDSB := SndObjGetFreeBuffer(hSO);

    if (not bWantSound) then
        exit;   // No sound our Work is done

    if (pDSB <> nil) then
    begin
        case ScreenX of
            320:
                pDSB^^.SetPan^(pDSB, trunc((20000.0 *
                    ((uobject^.dst.right + uobject^.dst.left) / 2) / 320.0) - 10000.0));
            640:
                pDSB^^.SetPan^(pDSB, trunc((20000.0 *
                    ((uobject^.dst.right + uobject^.dst.left) / 2) / 640.0) - 10000.0));
            1024:
                pDSB^^.SetPan^(pDSB, trunc((20000.0 *
                    ((uobject^.dst.right + uobject^.dst.left) / 2) / 1024.0) - 10000.0));
            1280:
                pDSB^^.SetPan^(pDSB, trunc((20000.0 *
                    ((uobject^.dst.right + uobject^.dst.left) / 2) / 1280.0) - 10000.0));
        end;

        pDSB^^.Play^(pDSB, 0, 0, 0);
    end;
end;
{$endif}

procedure UpdateDisplayList;
var
    this:          PDBlNode;
    save:          PDBlNode;
    thisTickCount: DWORD;
    tickDiff:      DWORD;
    maxx, maxy:    Double;
    maxframe:      Double;
    input:         DWORD := 0;
    event:         Boolean := FALSE;
begin
    thisTickCount := GetTickCount;
    tickDiff := thisTickCount - lastTickCount;
    if bTest then
        input := input or (KEY_RIGHT or KEY_FIRE);

    lastTickCount := thisTickCount;

    input := ReadGameInput;

    if (showDelay <> 0) then
    begin
        showDelay -:= Longint(tickDiff);
        if (showDelay < 0) then
        begin
            showDelay := 0;
            lastShield := FALSE;
            initShip( FALSE );
        end;
    end;

    // update the ship
    if ( showDelay = 0 ) then
    begin
        DL.posx +:= DL.velx * tickDiff;
        DL.posy +:= DL.vely * tickDiff;
    end;
    if( DL.posx > MAX_SHIP_X ) then
    begin
        DL.posx := MAX_SHIP_X;
        DL.velx := -DL.velx;
        event := TRUE;
    end else
    if ( DL.posx < 0 ) then
    begin
        DL.posx := 0;
        DL.velx := -DL.velx;
        event := TRUE;
    end;
    if ( DL.posy > MAX_SHIP_Y ) then
    begin
        DL.posy := MAX_SHIP_Y;
        DL.vely := -DL.vely;
        event := TRUE;
    end else
    if ( DL.posy < 0 ) then
    begin
        DL.posy := 0;
        DL.vely := -DL.vely;
        event := TRUE;
    end;
    if (event) then
    begin
    {$ifdef USE_DSOUND}
        if (bWantSound) then
        begin
            playPanned(hsoShipBounce, @DL);
        end;
    {$endif}
        event := FALSE;
    end;

    event := (showDelay <> 0) or ((input and KEY_SHIELD) = KEY_SHIELD);

    if (event <> lastShield) then
    begin
        if (event) and (showDelay = 0) then
        begin
        {$ifdef USE_DSOUND}
            if (bWantSound) then
            begin
                SndObjPlay(hsoShieldBuzz, DSBPLAY_LOOPING);
            end;
        {$endif}
            bPlayBuzz := TRUE;
        end else
        begin
        {$ifdef USE_DSOUND}
            if (bWantSound) then
            begin
                SndObjStop(hsoShieldBuzz);
            end;
        {$endif}
            bPlayBuzz := FALSE;
        end;
        lastShield := event;
    end;
    if (event) then
    begin
        input := input and (not KEY_FIRE);
    end;

    if (input and KEY_FIRE) <> 0 then
    begin
        if ( showDelay = 0 ) then
        begin
            // add a bullet to the scene
            score -:= 1;
            if (score < 0) then
                score := 0;
        {$ifdef USE_DSOUND}
            if (bWantSound) then
            begin
                SndObjPlay(hsoFireBullet, 0);
            end;
        {$endif}

            addObject( OBJ_BULLET, Dirx[trunc(DL.frame)]*6.0 + 16.0 + DL.posx,
                Diry[trunc(DL.frame)]*6.0 + 16.0 + DL.posy,
                Dirx[trunc(DL.frame)]*500.0/1000.0,
                Diry[trunc(DL.frame)]*500.0/1000.0 );
        end;
    end;

    event := FALSE;
    if ( input and KEY_LEFT ) <> 0 then
    begin
        DL.frame -:= 1.0;
        if ( DL.frame < 0.0 ) then
            DL.frame +:= MAX_SHIP_FRAME;
    end;
    if ( input and KEY_RIGHT ) <> 0 then
    begin
        DL.frame +:= 1.0;
        if ( DL.frame >= MAX_SHIP_FRAME) then
            DL.frame -:= MAX_SHIP_FRAME;
    end;
    if ( input and KEY_UP ) <> 0 then
    begin
        DL.velx +:= Dirx[trunc(DL.frame)] * 10.0/1000.0;
        DL.vely +:= Diry[trunc(DL.frame)] * 10.0/1000.0;
        event := TRUE;
    end;
    if ( input and KEY_DOWN ) <> 0 then
    begin
        DL.velx -:= Dirx[trunc(DL.frame)] * 10.0/1000.0;
        DL.vely -:= Diry[trunc(DL.frame)] * 10.0/1000.0;
        event := TRUE;
    end;

    if (event <> lastThrust) then
    begin
        if (event) then
        begin
            input := input and (not KEY_STOP);
        {$ifdef USE_DSOUND}
            if (bWantSound) then
            begin
                SndObjStop(hsoSkidToStop);
                SndObjPlay(hsoEngineRev, DSBPLAY_LOOPING);
            end;
        {$endif}
            bPlayRev := TRUE;
        end else
        begin
        {$ifdef USE_DSOUND}
            if (bWantSound) then
            begin
                SndObjStop(hsoEngineRev);
            end;
        {$endif}
            bPlayRev := FALSE;
        end;

        lastThrust := event;
    end;

    if ( input and KEY_STOP ) <> 0 then
    begin
    {$ifdef USE_DSOUND}
        if (bWantSound) then
        begin
            if (DL.velx <> 0) or (DL.vely <> 0) then
                playPanned(hsoSkidToStop, @DL);
        end;
    {$endif}
        DL.velx := 0;
        DL.vely := 0;
    end;

    this := DL.next;
    repeat
        this^.posx +:= this^.velx * tickDiff;
        this^.posy +:= this^.vely * tickDiff;
        this^.frame +:= this^.delay * tickDiff;
        case this^.objtype of
            OBJ_DONUT:
            begin
                maxx := MAX_DONUT_X;
                maxy := MAX_DONUT_Y;
                maxframe := MAX_DONUT_FRAME;
            end;
            OBJ_PYRAMID:
            begin
                maxx := MAX_PYRAMID_X;
                maxy := MAX_PYRAMID_Y;
                maxframe := MAX_PYRAMID_FRAME;
            end;
            OBJ_SPHERE:
            begin
                maxx := MAX_SPHERE_X;
                maxy := MAX_SPHERE_Y;
                maxframe := MAX_SPHERE_FRAME;
            end;
            OBJ_CUBE:
            begin
                maxx := MAX_CUBE_X;
                maxy := MAX_CUBE_Y;
                maxframe := MAX_CUBE_FRAME;
            end;
            OBJ_BULLET:
            begin
                maxx := MAX_BULLET_X;
                maxy := MAX_BULLET_Y;
                maxframe := MAX_BULLET_FRAME;
                if ( this^.frame >= MAX_BULLET_FRAME) then
                begin
                    save := this;
                    this := this^.next;
                    DeleteFromList( save );
                    continue;
                end;
            end;
        end;
        if ( this <> @DL ) then
        begin
            if( this^.posx > maxx ) then
            begin
                this^.posx := maxx;
                this^.velx := -this^.velx;
            end else
            if ( this^.posx < 0 ) then
            begin
                this^.posx :=0;
                this^.velx := -this^.velx;
            end;
            if ( this^.posy > maxy ) then
            begin
                this^.posy := maxy;
                this^.vely := -this^.vely;
            end else
            if ( this^.posy < 0 ) then
            begin
                this^.posy :=0;
                this^.vely := -this^.vely;
            end;
            if ( this^.frame >= maxframe ) then
            begin
                this^.frame -:= maxframe;
            end;
            this := this^.next;
        end;
    until( this = @DL );
end;

procedure CheckForHits;
var
    bullet, target, save: PDBlNode;
    frame, x, y, l, t:    Longint;
    hit:                  Boolean;
label
    cont;
begin
    // update screen rects
    target := @DL;
    repeat
        frame := trunc(target^.frame);
        case target^.objtype of
            OBJ_DONUT:
            begin
                target^.dst.left := trunc(target^.posx);
                target^.dst.top := trunc(target^.posy);
                target^.dst.right := target^.dst.left + 64;
                target^.dst.bottom := target^.dst.top + 64;
                target^.src.left := 64 * (frame mod 5);
                target^.src.top := 64 * (frame div 5);
                target^.src.right := target^.src.left + 64;
                target^.src.bottom := target^.src.top + 64;
            end;
            OBJ_PYRAMID:
            begin
                target^.dst.left := trunc(target^.posx);
                target^.dst.top := trunc(target^.posy);
                target^.dst.right := target^.dst.left + 32;
                target^.dst.bottom := target^.dst.top + 32;
                target^.src.left := 32 * (frame mod 10);
                target^.src.top := 32 * (frame div 10);
                target^.src.right := target^.src.left + 32;
                target^.src.bottom := target^.src.top + 32;
            end;
            OBJ_SPHERE:
            begin
                target^.dst.left := trunc(target^.posx);
                target^.dst.top := trunc(target^.posy);
                target^.dst.right := target^.dst.left + 16;
                target^.dst.bottom := target^.dst.top + 16;
                target^.src.left := 16 * (frame mod 20);
                target^.src.top := 16 * (frame div 20);
                target^.src.right := target^.src.left + 16;
                target^.src.bottom := target^.src.top + 16;
            end;
            OBJ_CUBE:
            begin
                target^.dst.left := trunc(target^.posx);
                target^.dst.top := trunc(target^.posy);
                target^.dst.right := target^.dst.left + 16;
                target^.dst.bottom := target^.dst.top + 16;
                target^.src.left := 16 * (frame mod 20);
                target^.src.top := 16 * (frame div 20);
                target^.src.right := target^.src.left + 16;
                target^.src.bottom := target^.src.top + 16;
            end;
            OBJ_SHIP:
            begin
                target^.dst.left := trunc(target^.posx);
                target^.dst.top := trunc(target^.posy);
                target^.dst.right := target^.dst.left + 32;
                target^.dst.bottom := target^.dst.top + 32;
                if ( lastShield ) then
                    target^.src.top := 32 * (frame div 10) + 128
                else
                    target^.src.top := 32 * (frame div 10);
                target^.src.left := 32 * (frame mod 10);
                target^.src.right := target^.src.left + 32;
                target^.src.bottom := target^.src.top + 32;
            end;
            OBJ_BULLET:
            begin
                frame := trunc(target^.frame) div 20 div 4;
                target^.dst.left := trunc(target^.posx);
                target^.dst.top := trunc(target^.posy);
                target^.dst.right := target^.dst.left + 3;
                target^.dst.bottom := target^.dst.top + 3;
                target^.src.left := BULLET_X + frame*4;
                target^.src.top := BULLET_Y;
                target^.src.right := target^.src.left + 3;
                target^.src.bottom := target^.src.top + 3;
            end;
        end;
        target := target^.next;
    until( target = @DL );

    bullet := @DL;
    repeat
        hit := FALSE;
        if (bullet^.objtype <> OBJ_BULLET) and (bullet <> @DL) then
        begin
            bullet := bullet^.next;
            continue;
        end;

        x := (bullet^.dst.left + bullet^.dst.right) div 2;
        y := (bullet^.dst.top + bullet^.dst.bottom) div 2;

        target := DL.next;
        while (target <> @DL) do
        begin
            if( ( target^.objtype <> OBJ_DONUT ) and
                ( target^.objtype <> OBJ_PYRAMID ) and
                ( target^.objtype <> OBJ_SPHERE ) and
                ( target^.objtype <> OBJ_CUBE ) )  then goto cont;

            if( (x >= target^.dst.left) and
                (x <  target^.dst.right) and
                (y >= target^.dst.top) and
                (y <  target^.dst.bottom) ) then
            begin
                if (bullet <> @DL) or (not lastShield) then
                begin
                    // the bullet hit the target
                    case target^.objtype  of
                        OBJ_DONUT:
                        begin
                        {$ifdef USE_DSOUND}
                            if (bWantSound) then
                            begin
                                playPanned(hsoDonutExplode, target);
                            end;
                        {$endif}
                            addObject( OBJ_PYRAMID, target^.dst.left,
                                target^.dst.top, -1.0, -1.0 );
                            addObject( OBJ_PYRAMID, target^.dst.left,
                                target^.dst.top, -1.0, -1.0 );
                            addObject( OBJ_PYRAMID, target^.dst.left,
                                target^.dst.top, -1.0, -1.0 );
                            score +:= 10;
                        end;
                        OBJ_PYRAMID:
                        begin
                        {$ifdef USE_DSOUND}
                            if (bWantSound) then
                            begin
                                playPanned(hsoPyramidExplode, target);
                            end;
                        {$endif}
                            addObject( OBJ_SPHERE, target^.dst.left,
                                target^.dst.top, -1.0, -1.0 );
                            addObject( OBJ_CUBE, target^.dst.left,
                                target^.dst.top, -1.0, -1.0 );
                            addObject( OBJ_CUBE, target^.dst.left,
                                target^.dst.top, -1.0, -1.0 );
                            score +:= 20;
                        end;
                        OBJ_CUBE:
                        begin
                        {$ifdef USE_DSOUND}
                            if (bWantSound) then
                            begin
                                playPanned(hsoCubeExplode, target);
                            end;
                        {$endif}
                            addObject( OBJ_SPHERE, target^.dst.left,
                                target^.dst.top, -1.0, -1.0 );
                            addObject( OBJ_SPHERE, target^.dst.left,
                                target^.dst.top, -1.0, -1.0 );
                            score +:= 40;
                        end;
                        OBJ_SPHERE:
                        begin
                        {$ifdef USE_DSOUND}
                            if (bWantSound) then
                            begin
                                playPanned(hsoSphereExplode, target);
                            end;
                        {$endif}
                            score +:= 20;
                        end;
                    end;

                    l := target^.dst.left;
                    t := target^.dst.top;
                    DeleteFromList( target );
                end;
                hit := TRUE;
            end;

            if ( hit ) then
            begin
                if( bullet = @DL ) then
                begin
                    hit := FALSE;
                    if (not lastShield) and (showDelay = 0) and (not bTest) then
                    begin
                    {$ifdef USE_DSOUND}
                        if (bWantSound) then
                        begin
                            playPanned(hsoShipExplode, bullet);
                        end;
                    {$endif}
                        score -:= 150;
                        if (score < 0) then
                            score := 0;

                        addObject( OBJ_SPHERE, l, t, -1.0, -1.0 );
                        addObject( OBJ_SPHERE, l, t, -1.0, -1.0 );
                        addObject( OBJ_SPHERE, l, t, -1.0, -1.0 );
                        addObject( OBJ_SPHERE, l, t, -1.0, -1.0 );
                        addObject( OBJ_BULLET, l, t,
                            randDouble( -0.5, 0.5 ), randDouble( -0.5, 0.5 ) );
                        addObject( OBJ_BULLET, l, t,
                            randDouble( -0.5, 0.5 ), randDouble( -0.5, 0.5 ) );
                        addObject( OBJ_BULLET, l, t,
                            randDouble( -0.5, 0.5 ), randDouble( -0.5, 0.5 ) );
                        addObject( OBJ_BULLET, l, t,
                            randDouble( -0.5, 0.5 ), randDouble( -0.5, 0.5 ) );
                        addObject( OBJ_BULLET, l, t,
                            randDouble( -0.5, 0.5 ), randDouble( -0.5, 0.5 ) );
                        addObject( OBJ_BULLET, l, t,
                            randDouble( -0.5, 0.5 ), randDouble( -0.5, 0.5 ) );
                        addObject( OBJ_BULLET, l, t,
                            randDouble( -0.5, 0.5 ), randDouble( -0.5, 0.5 ) );
                        addObject( OBJ_BULLET, l, t,
                            randDouble( -0.5, 0.5 ), randDouble( -0.5, 0.5 ) );
                        addObject( OBJ_BULLET, l, t,
                            randDouble( -0.5, 0.5 ), randDouble( -0.5, 0.5 ) );
                        addObject( OBJ_BULLET, l, t,
                            randDouble( -0.5, 0.5 ), randDouble( -0.5, 0.5 ) );
                        initShip(TRUE);
                    end;
                end;

                break;
            end;
        cont:
            target := target^.next;
        end;

        if ( hit ) then
        begin
            save := bullet;
            bullet := bullet^.next;
            DeleteFromList( save );
        end else
        begin
            bullet := bullet^.next;
        end;
    until (bullet = @DL);
end;

procedure DisplayFrameRate;
var
    time2: DWORD;
    buff:  array[0..255] of Char;
begin
    dwFrameCount +:= 1;
    time2 := timeGetTime - dwFrameTime;
    if ( time2 > 1000 ) then
    begin
        dwFrames := (dwFrameCount*1000) div time2;
        dwFrameTime := timeGetTime;
        dwFrameCount := 0;
    end;
    if ( dwFrames = 0 ) then
         exit;

    if (dwFrames <> dwFramesLast) then
        dwFramesLast := dwFrames;

    if( dwFrames > 99 ) then
        dwFrames := 99;

    buff[0] := char(dwFrames div 10 + Ord('0'));
    buff[1] := char(dwFrames mod 10 + Ord('0'));
    buff[2] := #0;
    bltScore(buff, ScreenX div 2-25, 10);
end;

procedure DrawDisplayList;
var
    this:     PDblNode;
    last:     PDblNode;
    ddrval:   HResult;
    scorebuf: array[0..10] of Byte;
    rem:      Longint;
begin
    // blt everything in reverse order if we are doing destination transparency
    // calculate score string
    scorebuf[0] := score div 10000000 + Ord('0');
    rem := score mod 10000000;
    scorebuf[1] := rem div 1000000 + Ord('0');
    rem := score mod 1000000;
    scorebuf[2] := rem div 100000 + Ord('0');
    rem := score mod 100000;
    scorebuf[3] := rem div 10000 + Ord('0');
    rem := score mod 10000;
    scorebuf[4] := rem div 1000 + Ord('0');
    rem := score mod 1000;
    scorebuf[5] := rem div 100 + Ord('0');
    rem := score mod 100;
    scorebuf[6] := rem div 10 + Ord('0');
    rem := score mod 10;
    scorebuf[7] := rem + Ord('0');
{$ifdef USE_DSOUND}
    if ( bSoundEnabled ) then
    begin
        scorebuf[8]  := 14 + Ord('0');
        scorebuf[9]  := 13 + Ord('0');
        scorebuf[10] := 0;
    end else
{$endif}
    scorebuf[8] := 0;

    EraseScreen;
    if ( dwTransType = DDBLTFAST_DESTCOLORKEY ) then
    begin
        bltScore(PChar(@scorebuf), 10, ScreenY-26);

        if ( bShowFrameCount )  then
            DisplayFrameRate;

        this := DL.next; // start with the topmost bitmap
        last := DL.next; // don't blt it twice

        if (showDelay <> 0) then
            last := @DL;
    end else
    begin
        this := @DL; // start with the bottommost bitmap (the ship)
        last := @DL; // don't blt it twice

        if (showDelay <> 0) then
            this := this^.prev;
    end;

    repeat
        while( TRUE ) do
        begin
            if ( this^.src.right - this^.src.left ) = 0 then
                break; // new objects don't have a src rect yet

            ddrval := lpBackBuffer^^.BltFast^( lpBackBuffer, this^.dst.left, this^.dst.top, this^.surf, this^.src, dwTransType );
            if ( ddrval = DD_OK ) then
            begin
                break;
            end;
            if ( ddrval = DDERR_SURFACELOST ) then
            begin
                if ( not RestoreSurfaces ) then
                    exit;
            end;
            if ( ddrval <> DDERR_WASSTILLDRAWING ) then
                exit;
        end;
        if ( dwTransType <> DDBLTFAST_DESTCOLORKEY ) then
        begin
            this := this^.prev;
        end else
        begin
            this := this^.next;
        end;
    until ( this = last );

    if ( dwTransType <> DDBLTFAST_DESTCOLORKEY ) then
    begin
        bltScore(PChar(@scorebuf), 10, ScreenY-26);

        if ( bShowFrameCount ) then
            DisplayFrameRate;
    end;

    FlipScreen;
end;

function isDisplayListEmpty: Boolean;
var
    ptr: PDblNode;
begin
    ptr := DL.next;
    while (ptr <> @DL) do
    begin
        if (ptr^.objtype <> OBJ_BULLET) then
        begin
            Result := FALSE;
            exit;
        end;
        ptr := ptr^.next;
    end;
    Result := TRUE;
end;

function randInt( low, high: Longint ): Longint;
var
    range, num: Longint;
begin
    range := high - low;
    num := random(32768) mod range;
    Result := ( num + low );
end;


procedure UpdateFrame;
begin
    if ProgramState = PS_BEGINREST then
    begin
    {$ifdef USE_DSOUND}
        if (bWantSound) then
        begin
            SndObjPlay(hsoBeginLevel, 0);
        end;
    {$endif}
        ProgramState := PS_REST;
        //
        // FALLTHRU
        //
    end;

    case ProgramState of
        PS_SPLASH:
        begin
            // display the splash screen
            bltSplash;
            exit;
        end;
        PS_ACTIVE:
        begin
            UpdateDisplayList;
            CheckForHits;
            DrawDisplayList;
            if ( isDisplayListEmpty ) then
            begin
            {$ifdef USE_DSOUND}
                if (bWantSound) then
                begin
                    SndObjStop(hsoEngineIdle);
                    SndObjStop(hsoEngineRev);
                end;
            {$endif}
                bPlayIdle := FALSE;
                bPlayRev := FALSE;
                lastThrust := lastShield = FALSE;
                ProgramState := PS_BEGINREST;
                restCount := GetTickCount;
                inc(Level);
                initLevel( level );
            end;
            exit;
        end;
        PS_REST:
        begin
            if ( ( GetTickCount - restCount ) > ShowLevelCount ) then
            begin
            {$ifdef USE_DSOUND}
            if (bWantSound) then
            begin
                 SndObjPlay(hsoEngineIdle, DSBPLAY_LOOPING);
            end;
            {$endif}
                bPlayIdle := TRUE;
                lastTickCount := GetTickCount;
                ProgramState := PS_ACTIVE;
            end else
            begin
                DisplayLevel;
            end;
            exit;
        end;
    end; // case
end;

procedure InitScreenParams;
begin
    MAX_SCREEN_X  := ScreenX-1;
    MAX_SCREEN_Y  := ScreenY-1;
    MAX_DONUT_X   := MAX_SCREEN_X - 64;
    MAX_DONUT_Y   := MAX_SCREEN_Y - 64;
    MAX_PYRAMID_X := MAX_SCREEN_X - 32;
    MAX_PYRAMID_Y := MAX_SCREEN_Y - 32;
    MAX_SPHERE_X  := MAX_SCREEN_X - 16;
    MAX_SPHERE_Y  := MAX_SCREEN_Y - 16;
    MAX_CUBE_X    := MAX_SCREEN_X - 16;
    MAX_CUBE_Y    := MAX_SCREEN_Y - 16;
    MAX_SHIP_X    := MAX_SCREEN_X - 32;
    MAX_SHIP_Y    := MAX_SCREEN_Y - 32;
    MAX_BULLET_X  := MAX_SCREEN_X - 3;
    MAX_BULLET_Y  := MAX_SCREEN_Y - 3;
end;

(*
 * WinMain
 *)
var
    msg: TMsg;
    i: Longint;
    lpCmdLine: PChar;
    pBuff: array[0..79] of Char;
begin
    for i := 1 to ParamCount do
    begin
        if ParamStr(i) = '-e' then
            bUseEmulation := TRUE
        else
        if ParamStr(i) = '-t' then
            bTest := TRUE
        else
        if ParamStr(i) = '-S' then
            bWantSound := TRUE
        else
        if ParamStr(i) = '-x' then
        begin
            bStress := TRUE;
            bTest := TRUE;
        end else
        begin
            StrPCopy(pBuff, ParamStr(i));
            lpCmdLine := pBuff;
            ScreenX   := getint(lpCmdLine, 640);
            ScreenY   := getint(lpCmdLine, 480);
            ScreenBpp := getint(lpCmdLine, 8);
        end;
    end;

    InitScreenParams;

    if ( not initApplication(CmdShow) ) then
        halt(0);

    if ( not InitializeGame ) then
    begin
        DestroyWindow( hWndMain );
        halt(0);
    end;

    dwFrameTime := timeGetTime;

    while( TRUE ) do
    begin
        if ( PeekMessage( msg, 0, 0, 0, PM_NOREMOVE ) ) then
        begin
            if( not GetMessage( msg, 0, 0, 0 ) ) then
            begin
                halt(msg.wParam);
            end;
            if ( TranslateAccelerator( hWndMain, hAccel, msg ) = 0 ) then
            begin
                TranslateMessage(msg);
                DispatchMessage(msg);
            end;
        end
        else
        if ( bIsActive ) then
        begin
            UpdateFrame;
        end else
        begin
            WaitMessage;
        end;
    end;

end. (* WinMain *)